"""
MIT License

Copyright (c) 2020-present shay (shayypy)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

------------------------------------------------------------------------------

This project includes code from https://github.com/Rapptz/discord.py, which is
available under the MIT license:

The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
import re
from typing import TYPE_CHECKING, Any, Dict, Optional, List, Tuple, Union

import guilded.abc  # type: ignore

from .asset import Asset
from .enums import ChannelType, FileType, MediaType
from .errors import InvalidArgument
from .file import File
#from .gateway import GuildedVoiceWebSocket
from .group import Group
from .message import ChatMessage, HasContentMixin
from .user import Member
from .utils import MISSING, get, ISO8601
from .status import Game

if TYPE_CHECKING:
    from .emoji import Emoji
    from .embed import Embed
    from .team import Team
    from .user import User
    from .webhook import Webhook


__all__ = (
    'Announcement',
    'AnnouncementChannel',
    'AnnouncementReply',
    'Availability',
    'ChatChannel',
    'DMChannel',
    'Doc',
    'DocsChannel',
    'DocReply',
    'ForumChannel',
    'ForumReply',
    'ForumTopic',
    'Media',
    'MediaChannel',
    'MediaReply',
    'ListChannel',
    'ListItem',
    'ListItemNote',
    'SchedulingChannel',
    'TextChannel',
    'Thread',
    'VoiceChannel',
)


class ChatChannel(guilded.abc.TeamChannel, guilded.abc.Messageable):
    """Represents a chat channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        self.type = ChannelType.chat
        self._channel_id = self.id

    async def create_webhook(self, *, name: str, avatar: Optional[Union[bytes, File]] = None) -> Webhook:
        """|coro|

        Create a webhook in this channel.

        Parameters
        -----------
        name: :class:`str`
            The webhook's name.
        avatar: Optional[Union[:class:`bytes`, :class:`File`]]
            A :term:`py:bytes-like object` or :class:`File` for the webhook's avatar.
            If the client is a bot user, providing this does nothing.
            Else, providing this causes the library to perform an extra API request.

        Returns
        --------
        :class:`Webhook`
            The created webhook.

        Raises
        -------
        HTTPException
            Creating the webhook failed.
        Forbidden
            You do not have permissions to create a webhook.
        """

        webhook = await self.team.create_webhook(
            channel=self,
            name=name,
            avatar=avatar,
        )
        return webhook

    async def webhooks(self) -> List[Webhook]:
        """|coro|

        Fetch the list of webhooks in this channel.

        .. warning::

            This endpoint cannot be paginated.

        Returns
        --------
        List[:class:`Webhook`]
            The webhooks in this channel.

        Raises
        -------
        Forbidden
            You do not have permissions to get the webhooks.
        """

        webhooks = await self.team.webhooks(channel=self)
        return webhooks

TextChannel = ChatChannel  # discord.py

class Doc(HasContentMixin):
    """Represents a doc in a :class:`DocsChannel`.

    .. container:: operations

        .. describe:: x == y

            Checks if two docs are equal.

        .. describe:: x != y

            Checks if two docs are not equal.

        .. describe:: str(x)

            Returns the title of the doc.

    Attributes
    -----------
    id: :class:`int`
        The doc's ID.
    title: :class:`str`
        The doc's title.
    content: :class:`str`
        The doc's text content.
    channel: :class:`.DocsChannel`
        The channel that the doc is in.
    group: :class:`.Group`
        The group that the doc is in.
    team: :class:`.Team`
        The team that the doc is in.
    tags: List[:class:`str`]
        The list of tags assigned to the doc.
    public: :class:`bool`
        Whether the doc is public.
    draft: :class:`bool`
        Whether the doc is a draft.
    created_at: :class:`datetime.datetime`
        When the doc was created.
    edited_at: Optional[:class:`datetime.datetime`]
        When the doc was last modified.
    game: Optional[:class:`.Game`]
        The game associated with the doc.
    """
    def __init__(self, *, state, data, channel, game=None):
        super().__init__()
        self._state = state

        self.channel = channel
        self.group = channel.group
        self.team = channel.team

        self.author_id: str = data.get('createdBy')
        self.edited_by_id: Optional[str] = data.get('updatedBy') or data.get('modifiedBy')

        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))
        self.edited_at: Optional[datetime.datetime] = ISO8601(data.get('updatedAt') or data.get('modifiedAt'))

        self.id: int = int(data['id'])
        self.title: str = data['title']

        if state.userbot:
            self.game: Optional[Game] = game or (Game(game_id=data.get('gameId')) if data.get('gameId') else None)
            self._replies = {}

            tags = data.get('tags')
            if tags and isinstance(tags, str):
                tags = tags.split(',')
            self.tags: List[str] = tags or []

            self.public: bool = data.get('isPublic', False)
            self.credentialed: bool = data.get('isCredentialed', False)
            self.draft: bool = data.get('isDraft', False)
            self.content: str = self._get_full_content(data['content'])

        else:
            self.content: str = data['content']

    def __repr__(self) -> str:
        return f'<Doc id={self.id!r} title={self.title!r} author={self.author!r} channel={self.channel!r}>'

    def __eq__(self, other) -> bool:
        return isinstance(other, Doc) and other.id == self.id

    def __str__(self) -> str:
        return self.title

    @property
    def team_id(self) -> str:
        return self.team.id

    @property
    def group_id(self) -> str:
        return self.group.id

    @property
    def channel_id(self) -> str:
        return self.channel.id

    @property
    def replies(self) -> List[DocReply]:
        """List[:class:`.DocReply`]: The list of cached replies to this doc."""
        return list(self._replies.values())

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that created the doc."""
        return self.team.get_member(self.author_id)

    @property
    def edited_by(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that last edited the doc."""
        return self.team.get_member(self.author_id)

    async def add_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Add a reaction to this doc.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to add.
        """
        if self._state.userbot:
            await self._state.add_content_reaction('doc', self.id, emoji.id)
        else:
            emoji_id: int = getattr(emoji, 'id', emoji)
            await self._state.add_reaction_emote(self.channel.id, self.id, emoji_id)

    async def remove_self_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Remove one of your reactions from this doc.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to remove.
        """
        if self._state.userbot:
            await self._state.remove_self_content_reaction('doc', self.id, emoji.id)
        else:
            emoji_id: int = getattr(emoji, 'id', emoji)
            await self._state.remove_reaction_emote(self.channel.id, self.id, emoji_id)

    async def delete(self) -> None:
        """|coro|

        Delete this doc.
        """
        await self._state.delete_doc(self.channel.id, self.id)

    async def edit(self, *content, **kwargs) -> Doc:
        """|coro|

        Edit this doc.

        Parameters
        -----------
        content: Any
            The content of the doc.
        title: str
            The title of the doc.
        game: Optional[:class:`.Game`]
            The game associated with the doc.
        draft: Optional[:class:`bool`]
            Whether the doc should be a draft.
        public: Optional[:class:`bool`]
            Whether the doc should be public.
        tags: List[:class:`str`]
            The list of tags for this doc. Overrides the existing list, if any.
            You can provide a maximum of 539 tags. You may include duplicates,
            however this behavior is forbidden in the desktop client.

        Returns
        --------
        :class:`.Doc`
            The updated doc.
        """

        title = kwargs.pop('title', self.title)

        # TODO: This may have issues with non-text content if no content is provided explicitly to this method
        if content is not None:
            content = tuple(content)
            if len(content) < 1:
                content = (self.content,)
            if not self._state.userbot:
                content = content[0]

        payload = {
            'title': title,
            'content': content,
        }

        if self._state.userbot:
            # We need to provide this or else Guilded will throw a 500
            payload['isDraft'] = kwargs.pop('draft', self.draft)

            try:
                payload['gameId'] = kwargs.pop('game', self.game).id
            except (KeyError, AttributeError):
                pass
            try:
                payload['isPublic'] = kwargs.pop('public')
            except KeyError:
                pass
            try:
                payload['tags'] = kwargs.pop('tags')
            except KeyError:
                pass

        data = await self._state.update_doc(self.channel.id, self.id, payload=payload)
        doc = Doc(data=data, channel=self.channel, game=self.game, state=self._state)
        return doc

    async def reply(self, *content, **kwargs) -> DocReply:
        """|coro|

        |onlyuserbot|

        Reply to this doc.

        Parameters
        -----------
        \*content: Any
            The content to create the reply with.
        reply_to: Optional[:class:`.DocReply`]
            An existing reply to reply to.

        Returns
        --------
        :class:`.DocReply`
            The created reply.
        """
        data = await self._state.create_content_reply('doc', self.team.id, self.id, content=content, reply_to=kwargs.get('reply_to'))
        reply = DocReply(data=data['reply'], parent=self, state=self._state)
        return reply

    def get_reply(self, id: int) -> Optional[DocReply]:
        """Optional[:class:`.DocReply`]: Get a reply by its ID."""
        return self._replies.get(id)

    async def fetch_replies(self) -> List[DocReply]:
        """|coro|

        |onlyuserbot|

        Fetch the replies to this doc.

        Returns
        --------
        List[:class:`.DocReply`]
        """
        replies = []
        data = await self._state.get_content_replies('doc', self.id)
        for reply_data in data:
            reply = DocReply(data=reply_data, parent=self, state=self._state)
            replies.append(reply)

        return replies

    async def fetch_reply(self, id: int) -> DocReply:
        """|coro|

        |onlyuserbot|

        Fetch a reply to this doc.

        Parameters
        -----------
        id: :class:`int`
            The ID of the reply.

        Returns
        --------
        :class:`.DocReply`
        """
        data = await self._state.get_content_reply('docs', self.channel.id, self.id, id)
        reply = DocReply(data=data['metadata']['reply'], parent=self, state=self._state)
        return reply

    async def move(self, to) -> None:
        """|coro|

        |onlyuserbot|

        Move this doc to another :class:`.DocsChannel`.

        Parameters
        -----------
        to: :class:`.DocsChannel`
            The channel to move this doc to.
        """
        await self._state.move_doc(self.channel.id, self.id, to.id)


class DocsChannel(guilded.abc.TeamChannel):
    """Represents a docs channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        if self._state.userbot:
            self.type = ChannelType.doc
            self._shareable_content_type = 'docs'
        else:
            self.type = ChannelType.docs

        self._docs = {}

    @property
    def docs(self) -> List[Doc]:
        """List[:class:`.Doc`]: The list of cached docs in this channel."""
        return list(self._docs.values())

    def get_doc(self, id: int) -> Optional[Doc]:
        """Optional[:class:`.Doc`]: Get a cached doc in this channel."""
        return self._docs.get(id)

    async def getch_doc(self, id: int) -> Doc:
        return self.get_doc(id) or await self.fetch_doc(id)

    async def create_doc(self, *content, **kwargs) -> Doc:
        """|coro|

        Create a new doc in this channel.

        Parameters
        -----------
        content: Any
            The content to create the doc with.
        title: :class:`str`
            The doc's title.
        game: Optional[:class:`.Game`]
            The game associated with the doc.
        draft: Optional[:class:`bool`]
            Whether the doc should be a draft.
        tags: List[:class:`str`]
            The list of tags for this doc.
            You can provide a maximum of 539 tags. You may include duplicates,
            however this behavior is forbidden in the desktop client.

        Returns
        --------
        :class:`.Doc`
            The created doc.
        """

        payload = {
            'title': kwargs.pop('title'),
            'content': content,
        }

        if self._state.userbot:
            # We need to provide this or else Guilded will throw a 500
            payload['isDraft'] = kwargs.pop('draft', False)

            try:
                payload['gameId'] = kwargs.pop('game').id
            except (KeyError, AttributeError):
                pass
            try:
                payload['tags'] = kwargs.pop('tags')
            except KeyError:
                pass

        data = await self._state.create_doc(self.id, payload=payload)
        doc = Doc(data=data, channel=self, game=kwargs.get('game'), state=self._state)
        return doc

    async def fetch_doc(self, id: int) -> Doc:
        """|coro|

        Fetch an doc in this channel.

        Parameters
        -----------
        id: :class:`int`
            The doc's ID.

        Returns
        --------
        :class:`.Doc`
        """
        data = await self._state.get_doc(self.id, id)
        doc = Doc(data=data, channel=self, state=self._state)
        return doc

    async def fetch_docs(self, *, limit: int = 50, before: datetime.datetime = None) -> List[Doc]:
        """|coro|

        Fetch multiple docs in this channel.

        All parameters are optional.

        Parameters
        -----------
        limit: :class:`int`
            The maximum number of docs to return. Defaults to 50.
        before: :class:`datetime.datetime`
            The latest date that an doc can be from. Defaults to the
            current time.

        Returns
        --------
        List[:class:`.Doc`]
        """
        if self._state.userbot:
            before = before or datetime.datetime.now(datetime.timezone.utc)
            data = await self._state.get_docs(self.id, limit=limit, before=before)
        else:
            data = await self._state.get_docs(self.id)
        docs = []
        for doc_data in data:
            docs.append(Doc(data=doc_data, channel=self, state=self._state))

        return docs


class ForumTopic(HasContentMixin):
    """Represents a forum topic.

    Attributes
    -----------
    id: :class:`int`
        The topic's ID.
    title: :class:`str`
        The topic's title.
    content: Optional[:class:`str`]
        The topic's content.
    channel: :class:`.ForumChannel`
        The forum channel that the topic is in.
    created_at: :class:`datetime.datetime`
        When the topic was created.
    bumped_at: Optional[:class:`datetime.datetime`]
        When the topic was last bumped. This may be the same as
        :attr:`.created_at`.
    edited_at: Optional[:class:`datetime.datetime`]
        When the topic was last edited.
    stickied: Optional[:class:`bool`]
        Whether the topic is stickied (pinned) in its channel.
    locked: Optional[:class:`bool`]
        Whether the topic is locked.
    deleted: Optional[:class:`bool`]
        Whether the topic is deleted.
    deleted_by: Optional[:class:`.Member`]
        Who deleted this topic. This will only be present through
        :meth:`on_forum_topic_delete`.
    reply_count: :class:`int`
        How many replies the topic has.
        If the client is a user account, this represents an API property, else
        this is the length of :attr:`.replies`.
    """

    __slots__: Tuple[str, ...] = (
        '_state',
        'channel',
        'channel_id',
        'team_id',
        'group_id',
        'id',
        'title',
        'content',
        'author_id',
        'webhook_id',
        'created_at',
        'edited_at',
        'bumped_at',
        'game_id',
        'visibility',
        'stickied',
        'locked',
        'shared',
        'deleted',
        'deleted_by',
        '_reply_count',
        '_replies',
    )

    def __init__(self, *, state, data, channel: ForumChannel):
        super().__init__()
        self._state = state
        self.channel = channel

        self.channel_id: str = data.get('channelId')
        self.team_id: str = data.get('teamId') or data.get('serverId')
        self.group_id: Optional[str] = data.get('groupId')

        self.id: int = data['id']
        self.title: str = data['title']
        self.content: Optional[str]
        if self._state.userbot:
            self.content = self._get_full_content(data['message'])
        else:
            self.content = data.get('content')

        self.author_id: str = data.get('createdBy')
        self.webhook_id: Optional[str] = data.get('createdByWebhookId')
        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))
        self.edited_at: Optional[datetime.datetime] = ISO8601(data.get('editedAt') or data.get('updatedAt'))
        self.bumped_at: Optional[datetime.datetime] = ISO8601(data.get('bumpedAt'))

        self.game_id: Optional[int] = data.get('gameId')
        self.visibility: Optional[str] = data.get('visibility')
        self.stickied: Optional[bool] = data.get('isSticky')
        self.locked: Optional[bool] = data.get('isLocked')
        self.shared: Optional[bool] = data.get('isShare')
        self.deleted: Optional[bool] = data.get('isDeleted')
        self.deleted_by: Optional[Member] = None

        self._reply_count: int = int(data.get('replyCount', 0))
        self._replies = {}

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return f'<ForumTopic id={self.id!r} title={self.title!r} channel={self.channel!r}>'

    @property
    def team(self) -> Team:
        """:class:`.Team`: The team that the topic is in."""
        return self._state._get_team(self.team_id) or self.channel.team

    @property
    def guild(self) -> Team:
        """:class:`.Team`: |dpyattr|

        This is an alias of :attr:`.team`.

        The team that the topic is in.
        """
        return self.team

    @property
    def server(self) -> Team:
        """:class:`.Team`: This is an alias of :attr:`.team`.

        The team that the topic is in.
        """
        return self.team

    @property
    def group(self) -> Group:
        """Optional[:class:`.Group`]: The group that the topic is in."""
        return self.team.get_group(self.group_id) or self.channel.group

    @property
    def game(self) -> Game:
        """Optional[:class:`.Game`]: The game that the topic is for."""
        return Game(game_id=self.game_id)

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that created the topic."""
        return self.team.get_member(self.author_id)

    @property
    def replies(self) -> List[ForumReply]:
        return list(self._replies.values())

    @property
    def reply_count(self) -> int:
        return self._reply_count if self._reply_count is not None else len(self.replies)

    @property
    def share_url(self) -> str:
        return f'{self.channel.share_url}/{self.id}'

    def get_reply(self, id) -> Optional[ForumReply]:
        """Optional[:class:`.ForumReply`]: Get a reply by its ID."""
        return self._replies.get(id)

    async def fetch_replies(self, *, limit=50) -> List[ForumReply]:
        """|coro|

        |onlyuserbot|

        Fetch the replies to this topic.

        Returns
        --------
        List[:class:`.ForumReply`]
        """
        replies = []
        data = await self._state.get_forum_topic_replies(self.channel.id, self.id, limit=limit)
        for reply_data in data.get('threadReplies', data) or []:
            reply = ForumReply(data=reply_data, parent=self, state=self._state)
            replies.append(reply)

        return replies

    async def fetch_reply(self, id: int) -> ForumReply:
        """|coro|

        |onlyuserbot|

        Fetch a reply to this topic.

        Returns
        --------
        :class:`.ForumReply`
        """
        data = await self._state.get_content_reply('forums', self.channel.id, self.id, id)
        reply = ForumReply(data=data['metadata']['reply'], parent=self, state=self._state)
        return reply

    async def reply(self, *content, **kwargs) -> int:
        """|coro|

        |onlyuserbot|

        Create a new reply to this topic.

        Parameters
        -----------
        content: Any
            The content to create the reply with.
        reply_to: Optional[:class:`.ForumReply`]
            An existing reply to reply to.

        Returns
        --------
        :class:`int`
            The ID of the created reply.

            .. note::
                Guilded does not return the full object in response to this.
                Nevertheless, if you are connected to the gateway, it should
                end up getting cached and accessible via :meth:`.get_reply`.
        """
        data = await self._state.create_forum_topic_reply(self.channel.id, self.id, content=content, reply_to=kwargs.get('reply_to'))
        return data['replyId']

    async def delete(self) -> None:
        """|coro|

        |onlyuserbot|

        Delete this topic.
        """
        await self._state.delete_forum_topic(self.channel.id, self.id)

    async def edit(self, *content, **kwargs) -> None:
        """|coro|

        |onlyuserbot|

        Edit this topic.

        Parameters
        -----------
        \*content: Any
            The topic's content.
        title: :class:`str`
            The topic's title.
        """

        payload = {
            'title': kwargs.pop('title', self.title),
            'content': content,
        }

        data = await self._state.update_forum_topic(self.channel.id, self.id, payload=payload)
        try:
            self.edited_at = ISO8601(data['editedAt'])
        except KeyError:
            pass

    async def move(self, to: ForumChannel) -> None:
        """|coro|

        |onlyuserbot|

        Move this topic to another :class:`.ForumChannel`.

        Parameters
        -----------
        to: :class:`.ForumChannel`
            The forum to move this topic to.
        """
        await self._state.move_forum_topic(self.channel.id, self.id, to.id)

    async def lock(self) -> None:
        """|coro|

        |onlyuserbot|

        Lock this topic.
        """
        await self._state.lock_forum_topic(self.channel.id, self.id)

    async def unlock(self) -> None:
        """|coro|

        |onlyuserbot|

        Unlock this topic.
        """
        await self._state.unlock_forum_topic(self.channel.id, self.id)

    async def sticky(self) -> None:
        """|coro|

        |onlyuserbot|

        Sticky (pin) this topic.
        """
        await self._state.sticky_forum_topic(self.channel.id, self.id)

    async def unsticky(self) -> None:
        """|coro|

        |onlyuserbot|

        Unsticky (unpin) this topic.
        """
        await self._state.unsticky_forum_topic(self.channel.id, self.id)

    async def pin(self) -> None:
        """|coro|

        |onlyuserbot|

        Pin (sticky) this topic. This is an alias of :meth:`.sticky`.
        """
        await self.sticky()

    async def unpin(self) -> None:
        """|coro|

        |onlyuserbot|

        Unpin (unsticky) this topic. This is an alias of :meth:`.unsticky`.
        """
        await self.unsticky()

    async def add_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Add a reaction to this topic.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to add.
        """
        if self._state.userbot:
            await self._state.add_content_reaction('post', self.id, emoji.id)
        else:
            emoji_id: int = getattr(emoji, 'id', emoji)
            await self._state.add_reaction_emote(self.channel.id, self.id, emoji_id)

    async def remove_self_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Remove one of your reactions from this topic.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to remove.
        """
        if self._state.userbot:
            await self._state.remove_self_content_reaction('post', self.id, emoji.id)
        else:
            emoji_id: int = getattr(emoji, 'id', emoji)
            await self._state.remove_reaction_emote(self.channel.id, self.id, emoji_id)


class ForumChannel(guilded.abc.TeamChannel):
    """Represents a forum channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        if self._state.userbot:
            self.type = ChannelType.forum
            self._shareable_content_type = 'forums'
        else:
            self.type = ChannelType.forums

        self._topics = {}

    @property
    def topics(self) -> List[ForumTopic]:
        """List[:class:`.ForumTopic`]: The list of topics in this forum."""
        return list(self._topics.values())

    def get_topic(self, id) -> Optional[ForumTopic]:
        """Optional[:class:`.ForumTopic`]: Get a topic by its ID."""
        return self._topics.get(id)

    async def create_topic(
        self,
        *pos_content: Optional[Union[str, Embed, File, Emoji]],
        content: Optional[str] = MISSING,
        title: str,
    ) -> ForumTopic:
        """|coro|

        Create a new topic in this forum.

        Parameters
        -----------
        \*pos_content: Union[:class:`str`, :class:`.Embed`, :class:`.File`, :class:`.Emoji`]
            An argument list of the topic's body content, passed in the order
            that each element should display.
            If the client is a bot account, only the first value is used as content.
            This parameter cannot be combined with ``content``.
        content: :class:`str`
            The text content to include in the body of the topic.
            This parameter exists so that text content can be passed using a keyword argument.
            This parameter cannot be combined with ``pos_content``.
        title: :class:`str`
            The title to create the topic with.

        Returns
        --------
        :class:`.ForumTopic`
            The topic that was created.
        """

        if content is not MISSING and pos_content:
            raise ValueError('Cannot provide both content and pos_content')

        if self._state.userbot:
            if content is not MISSING:
                pos_content = (content,)

            content = await self._state.process_list_content(pos_content)

            data = await self._state.create_forum_topic(
                self.id,
                title=title,
                message=self._state.compatible_content(content),
            )

        else:
            from .http import handle_message_parameters

            if pos_content:
                content = pos_content[0]

            params = handle_message_parameters(content=content)

            data = await self._state.create_forum_topic(
                self.id,
                title=title,
                content=params.payload['content'],
            )
            data = data['forumTopic']

        topic = ForumTopic(data=data, channel=self, state=self._state)
        return topic

    async def create_thread(self, *, name: str, content: Optional[str] = None, **kwargs) -> ForumTopic:
        """|coro|

        |dpyattr|

        Create a new topic in this forum.

        Parameters
        -----------
        name: :class:`str`
            The title to create the topic with.
        content: Optional[:class:`str`]
            The content to create the topic with.
        """
        return await self.create_topic(title=name, content=str(content))

    async def fetch_topic(self, id: int) -> ForumTopic:
        """|coro|

        |onlyuserbot|

        Fetch a topic from this forum.

        Parameters
        -----------
        id: :class:`int`
            The topic's ID.

        Returns
        --------
        :class:`.ForumTopic`
            The topic by its ID.
        """
        data = await self._state.get_forum_topic(self.id, id)
        topic = ForumTopic(data=data.get('thread', data), channel=self, state=self._state)
        return topic

    async def getch_topic(self, id: int) -> ForumTopic:
        return self.get_topic(id) or await self.fetch_topic(id)

    async def fetch_topics(self, *, limit: int = 50, page: int = 1, before: datetime.datetime = None) -> List[ForumTopic]:
        """|coro|

        |onlyuserbot|

        Fetch the topics in this forum.

        All parameters are optional.

        Parameters
        -----------
        limit: :class:`int`
            The maximum number of topics to return. Defaults to 50.
        before: :class:`datetime.datetime`
            The latest date that a topic can be from. Defaults to the current
            time.

        Returns
        --------
        List[:class:`.ForumTopic`]
            The topics in this forum.
        """
        before = before or datetime.datetime.now(datetime.timezone.utc)
        data = await self._state.get_forum_topics(self.id, limit=limit, page=page, before=before)
        topics = []
        for topic_data in data.get('threads', data):
            topic = ForumTopic(data=topic_data, channel=self, state=self._state)

        return topics


class VoiceChannel(guilded.abc.TeamChannel, guilded.abc.Messageable):
    """Represents a voice channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        self.type = ChannelType.voice
        self._channel_id = self.id
        self._ws = None

    #async def connect(self):
    #    state = self._state

    #    connection_info = await state.get_voice_connection_info(self.id)
    #    endpoint = connection_info['endpoint']
    #    token = connection_info['token']

    #    ws_build = GuildedVoiceWebSocket.build( loop=self.loop)
    #    gws = await asyncio.wait_for(ws_build, timeout=60)
    #    if type(gws) != GuildedVoiceWebSocket:
    #        self.dispatch('error', gws)
    #        return

    #    self._ws = gws
    #    self.dispatch('connect')

    #    lobby = await state.get_voice_lobby(endpoint, self.id)
    #    lobby_connection_data = await state.connect_to_voice_lobby(
    #        endpoint,
    #        self.id,
    #        rtp_capabilities=lobby['routerRtpCapabilities']
    #    )

    #    dtls_parameters = lobby_connection_data['sendTransportOptions']['dtlsParameters']
    #    # The client transforms the default "auto" to "server" and sends only
    #    # the fingerprint where algorithm is "sha-256"
    #    dtls_parameters['role'] = 'server'

    #    transport = await state.connect_to_voice_transport(
    #        endpoint,
    #        self.id,
    #        transport_id=lobby_connection_data['sendTransportOptions']['id'],
    #        dtls_parameters=dtls_parameters
    #    )


class Thread(guilded.abc.TeamChannel, guilded.abc.Messageable):
    """Represents a thread in a :class:`.Team`\.

    Attributes
    -----------
    initial_message_id: :class:`str`
        The ID of the initial message in this thread.
    """
    def __init__(self, **fields):
        super().__init__(**fields)
        data = fields.get('data') or fields.get('channel', {})
        self.type = ChannelType.thread

        self._message_count = data.get('messageCount') or 0
        self.initial_message_id = data.get('threadMessageId')
        self._initial_message = self._state._get_message(self.initial_message_id)
        # This is unlikely to not be None given the temporal nature of message cache

        self._participant_ids = []

        for user_id in data.get('userIds') or []:
            self._participant_ids.append(user_id)

        for member_data in data.get('participants') or []:
            if member_data.get('id'):
                self._participant_ids.append(member_data['id'])

    @property
    def message_count(self) -> int:
        """:class:`int`: The number of messages in the thread.

        This may be inaccurate if this object has existed for an extended
        period of time since it does not get updated by the library when new
        messages are sent within the thread.
        """
        return int(self._message_count)

    @property
    def initial_message(self) -> Optional[ChatMessage]:
        """Optional[:class:`.ChatMessage`]: The initial message in this thread.

        This may be ``None`` if the message was not cached when this object was
        created. In this case, you may fetch the message with :meth:`.fetch_initial_message`.
        """
        return self._initial_message

    @property
    def participants(self) -> List[Member]:
        """List[:class:`.Member`]: The cached list of participants in this thread."""
        return [self.team.get_member(member_id) for member_id in self._participant_ids]

    async def archive(self) -> None:
        """|coro|

        Archive this thread.
        """
        request = self._state.archive_team_thread(self.team_id, self.group_id, self.id)
        await request

    async def restore(self) -> None:
        """|coro|

        Restore this thread.
        """
        request = self._state.restore_team_thread(self.team_id, self.group_id, self.id)
        await request

    async def leave(self) -> None:
        """|coro|

        Leave this thread.
        """
        request = self._state.leave_thread(self.id)
        await request

    async def fetch_initial_message(self) -> ChatMessage:
        """|coro|

        Fetch the initial message in this thread. Sometimes this may be
        available via :attr:`.initial_message`, but it is unlikely when
        dealing with existing threads because it relies on message cache.

        This is equivalent to:

        .. code-block:: python3

            initial_message = await thread.fetch_message(thread.initial_message_id)

        Returns
        --------
        :class:`.ChatMessage`
            The initial message in the thread.
        """
        message = await self.fetch_message(self.initial_message_id)
        return message


class DMChannel(guilded.abc.Messageable):
    """Represents a private channel between users.

    Attributes
    -----------
    last_message: Optional[:class:`.ChatMessage`]
        The most recent message in the channel.
    recipient: Optional[:class:`.User`]
        The user on the opposing end of this conversation;
        The first participant who is not yourself.
    created_at: :class:`datetime.datetime`
        When the channel was created.
    updated_at: Optional[:class:`datetime.datetime`]
        When the channel was last updated.
    deleted_at: Optional[:class:`datetime.datetime`]
        When the channel was deleted.
    archived_at: Optional[:class:`datetime.datetime`]
        When the channel was archived.
    auto_archive_at: Optional[:class:`datetime.datetime`]
        When the channel will automatically archive.
    """
    def __init__(self, *, state, data):
        data = data.get('channel', data)
        super().__init__(state=state, data=data)
        self.type = ChannelType.dm
        self.team = None
        self.group = None

        self._users = {}
        self.recipient = None
        for user_data in data.get('users', []):
            user = self._state.create_user(data=user_data)
            if user:
                self._users[user.id] = user
                if user.id != self._state.my_id:
                    self.recipient = user

        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))
        self.updated_at: Optional[datetime.datetime] = ISO8601(data.get('updatedAt'))
        self.deleted_at: Optional[datetime.datetime] = ISO8601(data.get('deletedAt'))
        self.archived_at: Optional[datetime.datetime] = ISO8601(data.get('archivedAt'))
        self.auto_archive_at: Optional[datetime.datetime] = ISO8601(data.get('autoArchiveAt'))
        self.voice_participants: List[User] = data.get('voiceParticipants')
        self.last_message: Optional[ChatMessage] = None
        if data.get('lastMessage'):
            message_data = data.get('lastMessage')
            author = self._users.get(message_data.get('createdBy'))
            message = self._state.create_message(channel=self, data=message_data, author=author)
            self.last_message = message

    @property
    def share_url(self) -> str:
        """:class:`str`: The URL which can be used to navigate to this channel."""
        return f'https://guilded.gg/chat/{self.id}'

    @property
    def users(self) -> List[User]:
        """List[:class:`.User`]: The list of participants in this DM, including yourself."""
        return list(self._users.values())

    def __repr__(self) -> str:
        return f'<DMChannel id={self.id!r} recipient={self.recipient!r}>'

    async def hide(self) -> None:
        """|coro|

        Visually hide this DM channel in the client.

        The channel's content will still exist, and the channel can be
        re-fetched with :meth:`User.create_dm` on whichever :class:`User` this
        channel is associated with.
        """
        await self._state.hide_dm_channel(self.id)


class Announcement(HasContentMixin):
    """Represents an announcement in an :class:`.AnnouncementChannel`.

    Attributes
    -----------
    id: :class:`str`
        The announcement's ID.
    title: :class:`str`
        The announcement's title.
    content: :class:`str`
        The announcement's text content.
    channel: :class:`.AnnouncementChannel`
        The channel that the announcement is in.
    group: :class:`.Group`
        The group that the announcement is in.
    team: :class:`.Team`
        The team that the announcement is in.
    public: :class:`bool`
        Whether the announcement is public.
    pinned: :class:`bool`
        Whether the announcement is pinned.
    created_at: :class:`datetime.datetime`
        When the announcement was created.
    edited_at: Optional[:class:`datetime.datetime`]
        When the announcement was last edited.
    slug: Optional[:class:`str`]
        The announcement's URL slug.
    game: Optional[:class:`.Game`]
        The game associated with the announcement.
    """
    def __init__(self, *, state, data, channel, game=None):
        super().__init__()
        self._state = state
        self.channel = channel
        self.group = channel.group
        self.team = channel.team
        self.game: Optional[Game] = game or (Game(game_id=data.get('gameId')) if data.get('gameId') else None)
        self.tags: str = data.get('tags')
        self._replies = {}

        for reply_data in data.get('replies', []):
            reply = AnnouncementReply(data=reply_data, parent=self, state=self._state)
            self._replies[reply.id] = reply

        self.public: bool = data.get('isPublic', False)
        self.pinned: bool = data.get('isPinned', False)
        self.slug: Optional[str] = data.get('slug')

        self.author_id: str = data.get('createdBy')
        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))
        self.edited_at: Optional[datetime.datetime] = ISO8601(data.get('editedAt'))

        self.id: str = data['id']
        self.title: str = data['title']
        self.content: str = self._get_full_content(data['content'])

    def __repr__(self):
        return f'<Announcement id={self.id!r} title={self.title!r} author={self.author!r}>'

    @property
    def team_id(self) -> str:
        return self.team.id

    @property
    def group_id(self) -> str:
        return self.group.id

    @property
    def channel_id(self) -> str:
        return self.channel.id

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that created the announcement."""
        return self.team.get_member(self.author_id)

    @property
    def blog_url(self) -> Optional[str]:
        if self.channel.vanity_url and self.slug:
            return f'{self.channel.vanity_url}/{self.slug}'
        return None

    @property
    def share_url(self) -> str:
        return f'{self.channel.share_url}/{self.id}'

    @property
    def replies(self) -> List[AnnouncementReply]:
        return list(self._replies.values())

    def get_reply(self, id) -> Optional[AnnouncementReply]:
        """Optional[:class:`.AnnouncementReply`]: Get a reply by its ID."""
        return self._replies.get(id)

    async def sticky(self) -> None:
        """|coro|

        Sticky (pin) this announcement.
        """
        await self._state.toggle_announcement_pin(self.channel.id, self.id, pinned=True)
        self.pinned = True

    async def unsticky(self) -> None:
        """|coro|

        Unsticky (unpin) this announcement.
        """
        await self._state.toggle_announcement_pin(self.channel.id, self.id, pinned=False)
        self.pinned = False

    async def pin(self) -> None:
        """|coro|

        Pin (sticky) this announcement. This is an alias of :meth:`.sticky`.
        """
        await self.sticky()

    async def unpin(self) -> None:
        """|coro|

        Unpin (unsticky) this announcement. This is an alias of :meth:`.unsticky`.
        """
        await self.unsticky()

    async def delete(self) -> None:
        """|coro|

        Delete this announcement.
        """
        await self._state.delete_announcement(self.channel.id, self.id)

    async def edit(self, *content, **kwargs) -> None:
        """|coro|

        Edit this announcement.

        Parameters
        -----------
        \*content: Any
            The content of the announcement.
        title: :class:`str`
            The title of the announcement.
        """

        payload = {
            'title': kwargs.pop('title', self.title),
            'content': content,
        }

        await self._state.update_announcement(self.channel.id, self.id, payload=payload)

    async def add_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Add a reaction to this announcement.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to add.
        """
        await self._state.add_content_reaction(self.channel.type.value, self.id, emoji.id)

    async def remove_self_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Remove one of your reactions from this announcement.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to remove.
        """
        if self._state.userbot:
            await self._state.remove_self_content_reaction(self.channel.type.value, self.id, emoji.id)
        else:
            emoji_id: int = getattr(emoji, 'id', emoji)
            await self._state.remove_reaction_emote(self.channel.id, self.id, emoji_id)

    async def fetch_replies(self) -> None:
        """|coro|

        Fetch the replies to this announcement.

        Returns
        --------
        List[:class:`.AnnouncementReply`]
        """
        replies = []
        data = await self._state.get_content_replies(self.channel.type.value, self.id)
        for reply_data in data:
            reply = AnnouncementReply(data=reply_data, parent=self, state=self._state)
            replies.append(reply)

        return replies

    async def fetch_reply(self, id: int) -> AnnouncementReply:
        """|coro|

        Fetch a reply to this announcement.

        Parameters
        -----------
        id: :class:`int`
            The ID of the reply.

        Returns
        --------
        :class:`.AnnouncementReply`
        """
        data = await self._state.get_content_reply('announcements', self.channel.id, self.id, id)
        reply = AnnouncementReply(data=data['metadata']['reply'], parent=self, state=self._state)
        return reply

    async def reply(self, *content, **kwargs) -> AnnouncementReply:
        """|coro|

        Reply to this announcement.

        Parameters
        -----------
        content: Any
            The content to create the reply with.
        reply_to: Optional[:class:`.AnnouncementReply`]
            An existing reply to reply to.

        Returns
        --------
        :class:`.AnnouncementReply`
            The created reply.
        """
        data = await self._state.create_content_reply(self.channel.type.value, self.team.id, self.id, content=content, reply_to=kwargs.get('reply_to'))
        reply = AnnouncementReply(data=data['reply'], parent=self, state=self._state)
        return reply


class AnnouncementChannel(guilded.abc.TeamChannel):
    """Represents an announcement channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        if self._state.userbot:
            self.type = ChannelType.announcement
            self._shareable_content_type = 'announcements'
        else:
            self.type = ChannelType.announcements

        self._announcements = {}

    @property
    def announcements(self) -> List[Announcement]:
        """List[:class:`.Announcement`]: The list of cached announcements in this channel."""
        return list(self._announcements.values())

    def get_announcement(self, id) -> Optional[Announcement]:
        """Optional[:class:`.Announcement`]: Get a cached announcement in this channel."""
        return self._announcements.get(id)

    async def getch_announcement(self, id: str) -> Announcement:
        return self.get_announcement(id) or await self.fetch_announcement(id)

    async def fetch_announcement(self, id: str) -> Announcement:
        """|coro|

        Fetch an announcement in this channel.

        Parameters
        -----------
        id: :class:`str`
            The announcement's ID.

        Returns
        --------
        :class:`.Announcement`
        """
        data = await self._state.get_announcement(self.id, id)
        announcement = Announcement(data=data['announcement'], channel=self, state=self._state)
        return announcement

    async def fetch_announcements(self, *, limit: int = 50, before: datetime.datetime = None) -> List[Announcement]:
        """|coro|

        Fetch multiple announcements in this channel.

        All parameters are optional.

        Parameters
        -----------
        limit: :class:`int`
            The maximum number of announcements to return. Defaults to 50.
        before: :class:`datetime.datetime`
            The latest date that an announcement can be from. Defaults to the
            current time.

        Returns
        --------
        List[:class:`.Announcement`]
        """
        before = before or datetime.datetime.now(datetime.timezone.utc)
        data = await self._state.get_announcements(self.id, limit=limit, before=before)
        announcements = []
        for announcement_data in data['announcements']:
            announcements.append(Announcement(data=announcement_data, channel=self, state=self._state))

        return announcements

    async def fetch_pinned_announcements(self) -> List[Announcement]:
        """|coro|

        Fetch all pinned announcements in this channel.

        Returns
        --------
        List[:class:`.Announcement`]
        """
        data = await self._state.get_pinned_announcements(self.id)
        announcements = []
        for announcement_data in data['announcements']:
            announcements.append(Announcement(data=announcement_data, channel=self, state=self._state))

        return announcements

    async def create_announcement(self, *content, **kwargs) -> Announcement:
        """|coro|

        Create an announcement in this channel.

        Parameters
        -----------
        content: Any
            The content of the announcement.
        title: :class:`str`
            The title of the announcement.
        game: Optional[:class:`.Game`]
            The game to be associated with this announcement.
        send_notifications: Optional[:class:`bool`]
            Whether to send notifications to all members ("Notify all
            members" in the client). Defaults to ``True`` if not specified.

        Returns
        --------
        :class:`.Announcement`
            The created announcement.
        """
        title = kwargs.pop('title')
        game = kwargs.pop('game', None)
        dont_send_notifications = not kwargs.pop('send_notifications', True)

        data = await self._state.create_announcement(
            self.id,
            title=title,
            content=content,
            game_id=(game.id if game else None),
            dont_send_notifications=dont_send_notifications
        )
        announcement = Announcement(data=data['announcement'], channel=self, game=game, state=self._state)
        return announcement


class Media(HasContentMixin):
    """Represents a media post in a :class:`.MediaChannel`.

    .. container:: operations

        .. describe:: x == y

            Checks if two medias are equal.

        .. describe:: x != y

            Checks if two medias are not equal.

        .. describe:: str(x)

            Returns the URL of the media.

        .. describe:: len(x)

            Returns the length of the media's URL.

    Attributes
    -----------
    id: :class:`int`
        The media's ID.
    title: :class:`str`
        The media's title.
    description: :class:`str`
        The media's description.
    url: :class:`str`
        The media's URL on Guilded's CDN.
    thumbnail: Optional[:class:`.Asset`]
        An asset for the media's thumbnail.
    channel: :class:`.MediaChannel`
        The channel that the media is in.
    group: :class:`.Group`
        The group that the media is in.
    team: :class:`.Team`
        The team that the media is in.
    public: :class:`bool`
        Whether the media is public.
    created_at: :class:`datetime.datetime`
        When the media was created.
    reply_count: :class:`int`
        How many replies the media has.
    game: Optional[:class:`.Game`]
        The game associated with the media.
    """
    def __init__(self, *, state, data, channel, game=None):
        super().__init__()
        self._state = state
        self.type = getattr(FileType, (data.get('type', 'image')), None)
        self.channel = channel
        self.group = channel.group
        self.team = channel.team
        self.game: Optional[Game] = game or (Game(game_id=data.get('gameId')) if data.get('gameId') else None)
        self.tags: List[str] = list(data.get('tags') or [])  # sometimes an empty string is present instead of a list
        self._replies = {}

        self.public: bool = data.get('isPublic', False)
        self.url: str = data.get('src')

        thumbnail = None
        if data.get('srcThumbnail'):
            thumbnail = Asset._from_media_thumbnail(state, data['srcThumbnail'])
        self.thumbnail: Optional[Asset] = thumbnail

        self.author_id: str = data.get('createdBy')
        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))

        self.id: int = int(data['id'])
        self.title: str = data['title']
        self.description: str = data.get('description', '')

        self.reply_count: int = int(data.get('replyCount', 0))

        if data.get('additionalInfo', {}).get('externalVideoSrc'):
            self.youtube_embed_url = data['additionalInfo']['externalVideoSrc']
            self.youtube_video_id = re.sub(r'^https?:\/\/(www\.)youtube\.com\/embed\/', '', self.youtube_embed_url)

    def __repr__(self) -> str:
        return f'<Media id={self.id!r} title={self.title!r} author={self.author!r}>'

    def __str__(self) -> str:
        return self.url

    def __len__(self) -> int:
        return len(str(self))

    def __eq__(self, other) -> bool:
        return isinstance(other, Media) and self.id == other.id and self.url == other.url

    def _update(self, data) -> None:
        try:
            self.title = data['title']
        except KeyError:
            pass
        try:
            self.description = data['description']
        except KeyError:
            pass
        try:
            self.tags = data['tags']
        except KeyError:
            pass
        try:
            self.url = data['src']
        except KeyError:
            pass
        try:
            self.type = getattr(FileType, data['type'])
        except (KeyError, AttributeError):
            pass
        try:
            self.game = Game(game_id=data['gameId'])
        except KeyError:
            pass

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The member that created this media."""
        return self.team.get_member(self.author_id)

    async def add_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Add a reaction to this media post.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to add.
        """
        await self._state.add_content_reaction(self.channel.content_type, self.id, emoji.id)

    async def remove_self_reaction(self, emoji: Emoji) -> None:
        """|coro|

        Remove one of your reactions from this media post.

        Parameters
        -----------
        emoji: :class:`.Emoji`
            The emoji to remove.
        """
        if self._state.userbot:
            await self._state.remove_self_content_reaction(self.channel.content_type, self.id, emoji.id)
        else:
            emoji_id: int = getattr(emoji, 'id', emoji)
            await self._state.remove_reaction_emote(self.channel.id, self.id, emoji_id)

    async def reply(self, *content, **kwargs) -> MediaReply:
        """|coro|

        Reply to this media.

        Parameters
        -----------
        content: Any
            The content to create the reply with.
        reply_to: Optional[:class:`.MediaReply`]
            An existing reply to reply to.

        Returns
        --------
        :class:`.MediaReply`
            The created reply.
        """
        data = await self._state.create_content_reply(self.channel.content_type, self.team.id, self.id, content=content, reply_to=kwargs.get('reply_to'))
        reply = MediaReply(data=data['reply'], parent=self, state=self._state)
        return reply

    def get_reply(self, id: int):
        """Optional[:class:`.MediaReply`]: Get a reply by its ID."""
        return self._replies.get(id)

    async def fetch_replies(self) -> List[MediaReply]:
        """|coro|

        |onlyuserbot|

        Fetch the replies to this media post.

        Returns
        --------
        List[:class:`.MediaReply`]
        """
        replies = []
        data = await self._state.get_content_replies(self.channel.content_type, self.id)
        for reply_data in data:
            reply = MediaReply(data=reply_data, parent=self, state=self._state)
            replies.append(reply)

        return replies

    async def fetch_reply(self, id: int) -> MediaReply:
        """|coro|

        Fetch a reply to this media.

        Parameters
        -----------
        id: :class:`int`
            The ID of the reply.

        Returns
        --------
        :class:`.MediaReply`
        """
        data = await self._state.get_content_reply(self.channel.type.value, self.channel.id, self.id, id)
        # metadata uses 'media' and not 'team_media'
        reply = MediaReply(data=data['metadata']['reply'], parent=self, state=self._state)
        return reply

    async def move(self, to: MediaChannel) -> None:
        """|coro|

        Move this media post to another :class:`.MediaChannel`.

        Parameters
        -----------
        to: :class:`.MediaChannel`
            The media channel to move this media post to.
        """
        await self._state.move_media(self.channel.id, self.id, to.id)

    async def delete(self) -> None:
        """|coro|

        Delete this media post.
        """
        await self._state.delete_media(self.channel.id, self.id)

    async def edit(self, *,
        title: str = None,
        description: str = None,
        file: Optional[File] = None,
        youtube_url: Optional[str] = None,
        tags: List[str] = None,
        game: Optional[Game] = None,
    ) -> Media:
        """|coro|

        Edit this media post.

        All parameters are optional.

        Parameters
        -----------
        title: :class:`str`
            The title of the media.
        description: :class:`str`
            The description of the media. Does not accept markdown or any
            inline content.
        file: :class:`.File`
            The file to upload.
        youtube_url: :class:`str`
            The YouTube embed URL to use (``https://www.youtube.com/embed/...``).
        game: :class:`.Game`
            The game associated with the media.
        tags: List[:class:`str`]
            The tags on the media.

        Returns
        --------
        :class:`.Media`
            The newly updated media.
        """

        if file and youtube_url:
            raise ValueError('Must not specify both file and youtube_url')

        payload = {
            'id': self.id,
            'title': title or self.title,
            'description': description or self.description,
        }

        if tags is not None:
            payload['tags'] = tags
        else:
            payload['tags'] = self.tags

        if file:
            file.set_media_type(MediaType.media_channel_upload)
            await file._upload(self._state)
            payload['src'] = file.url
            payload['type'] = file.file_type.value
        elif youtube_url:
            data = await self._state.upload_third_party_media(youtube_url)
            payload['src'] = data['url']
            payload['additionalInfo'] = {'externalVideoSrc': youtube_url}
            payload['type'] = FileType.video.value
        else:
            payload['src'] = self.url
            payload['type'] = self.type.value

        if game is not None:
            payload['gameId'] = game.id
        elif self.game:
            payload['gameId'] = self.game.id

        await self._state.create_media(self.channel.id, payload=payload)

        self._update(payload)
        return self

    async def read(self) -> bytes:
        """|coro|

        Fetches the raw data of this media as a :class:`bytes`.

        Returns
        --------
        :class:`bytes`
            The raw data of this media.
        """
        return await self._state.read_filelike_data(self)


class ListItemNote(HasContentMixin):
    """Represents the note on a :class:`.ListItem`\.

    Attributes
    -----------
    parent: :class:`.ListItem`
        The note's parent item.
    content: Optional[:class:`str`]
        The note's content.
    """
    def __init__(self, *, data, parent: ListItem):
        super().__init__()
        self._state = parent._state

        self.parent = parent

        self.content: Optional[str]
        if self._state.userbot:
            self.content = self._get_full_content(data)
        else:
            self.content = data.get('content')
            self._mentions = self._create_mentions(data.get('mentions'))

    def __repr__(self) -> str:
        return f'<ListItemNote parent={self.parent!r} author={self.author!r}>'

    @property
    def team_id(self) -> str:
        return self.parent.team_id

    @property
    def team(self) -> Team:
        return self.parent.team

    @property
    def group_id(self) -> str:
        return self.parent.group_id

    @property
    def channel_id(self) -> str:
        return self.parent.channel_id

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that created the note."""
        return self.parent.team.get_member(self.parent._note_author_id)

    @property
    def updated_by(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that last updated the note."""
        return self.parent.team.get_member(self.parent._note_updated_by_id)

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the note was last updated."""
        return self.parent._note_updated_at

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: When the note was created."""
        return self.parent._note_created_at

    async def delete(self) -> None:
        """|coro|

        Delete this note.
        """
        await self.edit(content=None)

    async def edit(
        self,
        *pos_content: Optional[Union[str, Embed, File, Emoji]],
        content: Optional[str] = MISSING,
    ) -> ListItemNote:
        """|coro|

        Edit this note.

        Parameters
        -----------
        \*pos_content: Union[:class:`str`, :class:`.Embed`, :class:`.File`, :class:`.Emoji`]
            An argument list of the list's main content, passed in the order
            that each element should display.
            If the client is a bot account, only the first value is used as content.
            This parameter cannot be combined with ``content``.
        content: :class:`str`
            The text content to include as the main content of the item.
            This parameter exists so that text content can be passed using a keyword argument.
            This parameter cannot be combined with ``pos_content``.

        Returns
        --------
        :class:`.ListItemNote`
            The newly edited note.
        """

        if content is not MISSING and pos_content:
            raise ValueError('Cannot provide both content and pos_content')

        if self._state.userbot:
            if content is not MISSING:
                pos_content = (content,)

            if pos_content == (None,):
                payload = {
                    'note': None,
                }
            else:
                payload = {
                    'note': await self._state.process_list_content(pos_content),
                }
            data = await self._state.edit_list_item_message(
                self.channel_id,
                self.parent.id,
                payload=payload,
            )
            item = ListItem(data=data, channel=self.parent.channel, state=self._state)
            return item.note

        else:
            if pos_content:
                content = pos_content[0]

            payload = {
                'message': self.parent.message,
            }
            if content is not MISSING and content is not None:
                payload['note'] = {
                    'content': content,
                }

            data = await self._state.update_list_item(
                self.channel_id,
                self.parent.id,
                payload=payload,
            )
            item = ListItem(data=data['listItem'], channel=self.parent.channel, state=self._state)
            return item.note


class ListItem(HasContentMixin):
    """Represents an item in a :class:`.ListChannel`.

    Attributes
    -----------
    id: :class:`str`
        The item's ID.
    channel: :class:`.ListChannel`
        The channel that the item is in.
    group: :class:`.Group`
        The group that the item is in.
    team: :class:`.Team`
        The team that the item is in.
    created_at: :class:`datetime.datetime`
        When the item was created.
    message: :class:`str`
        The main message of the item.
    position: :class:`int`
        Where the item is in its :attr:`.channel`. A value of ``0`` is
        at the bottom of the list visually.
    note: :class:`.ListItemNote`
        The item's note.
    updated_at: Optional[:class:`datetime.datetime`]
        When the item was last updated.
    completed_at: Optional[:class:`datetime.datetime`]
        When the item was marked as completed.
    deleted_at: Optional[:class:`datetime.datetime`]
        When the item was deleted.
    """

    def __init__(self, *, state, data, channel: ListChannel):
        super().__init__()
        self._state = state
        self.channel = channel
        self.group = channel.group
        self.team = channel.team

        self.parent_id: Optional[str] = data.get('parentId') or data.get('parentListItemId')
        self.channel_id: str = data.get('channelId')
        self.team_id: str = data.get('teamId')
        self.group_id: Optional[str] = data.get('groupId')

        self.webhook_id: Optional[str] = data.get('webhookId') or data.get('createdByWebhookId')
        self.bot_id: Optional[str] = data.get('botId')
        self.author_id: str = data.get('createdBy')
        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))
        self.updated_by_id: Optional[str] = data.get('updatedBy')
        self.updated_at: Optional[datetime.datetime] = ISO8601(data.get('updatedAt'))
        self.completed_by_id: Optional[str] = data.get('completedBy')
        self.completed_at: Optional[datetime.datetime] = ISO8601(data.get('completedAt'))
        self._deleted_by: Optional[Member] = None
        self.deleted_by_id: Optional[str] = data.get('deletedBy')
        self.deleted_at: Optional[datetime.datetime] = ISO8601(data.get('deletedAt'))
        self._assigned_to: List[Dict[str, str]] = data.get('assignedTo') or []
        self._has_note: bool = data.get('hasNote')

        self.id: str = data['id']
        self.position: int = data.get('priority')
        self.message: str
        if self._state.userbot:
            self.message = self._get_full_content(data['message'])
        else:
            self.message = data['message']
            self._mentions = self._create_mentions(data.get('mentions'))

        _note: Dict[str, Any] = data.get('note') or {}
        self._note_author_id: Optional[str] = data.get('noteCreatedBy') or _note.get('createdBy')
        self._note_created_at: Optional[datetime.datetime] = ISO8601(data.get('noteCreatedAt') or _note.get('createdAt'))
        self._note_updated_by_id: Optional[str] = data.get('noteUpdatedBy') or _note.get('updatedBy')
        self._note_updated_at: Optional[datetime.datetime] = ISO8601(data.get('noteUpdatedAt') or _note.get('updatedAt'))
        self.note = ListItemNote(data=_note, parent=self)

    def __repr__(self) -> str:
        return f'<ListItem id={self.id!r} author={self.author!r} channel={self.team!r}>'

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that created the item."""
        return self.team.get_member(self.author_id)

    @property
    def updated_by(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that last updated the item."""
        return self.team.get_member(self.updated_by_id)

    @property
    def completed_by(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that marked the the item as completed."""
        return self.team.get_member(self.completed_by_id)

    @property
    def deleted_by(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The :class:`.Member` that deleted the item."""
        return self._deleted_by or self.team.get_member(self.deleted_by_id)

    @property
    def share_url(self) -> Optional[str]:
        return f'{self.channel.share_url}?listItemId={self.id}'

    @property
    def assigned_to(self) -> List[Member]:
        """List[:class:`.Member`]: The members that the item is assigned to,
        designated by the mentions in :attr:`.message`\."""

        if self._state.userbot:
            members = set()
            for assigned in self._assigned_to:
                id_ = assigned.get('mentionId')
                if assigned.get('mentionType') == 'person':
                    member = self.team.get_member(id_)
                    if member:
                        members.add(member)
                elif assigned.get('mentionType') == 'role':
                    role = self.team.get_role(int(id_))
                    if role:
                        for member in role.members:
                            members.add(member)
        else:
            members = set(self.user_mentions)
            for role in self.role_mentions:
                for member in role.members:
                    members.add(member)

        return list(members)

    @property
    def parent(self) -> Optional[ListItem]:
        """Optional[:class:`.ListItem`]: The item that this item is a child of."""
        return self.channel.get_item(self.parent_id)

    def has_note(self) -> bool:
        """:class:`bool`: Whether the item has a note."""
        # This property (hasNote) is not returned in the gateway
        # event for list items being created
        if self._has_note is not None:
            return self._has_note
        return self.note.content is not None

    async def fetch_parent(self) -> ListItem:
        """|coro|
        
        Fetch the item that this item is a child of, if it exists.

        Returns
        --------
        :class:`.ListItem`

        Raises
        -------
        ValueError
            This item has no parent.
        """
        if self.parent_id is None:
            raise ValueError('This item has no parent.')

        return await self.channel.fetch_item(self.parent_id)

    async def fetch_note(self) -> ListItemNote:
        """|coro|

        Fetch this item's note. This should only be necessary if you obtained
        this object through :meth:`ListChannel.fetch_items`.

        Returns
        --------
        :class:`.ListItemNote`
        """
        item = await self.channel.fetch_item(self.id)
        self.note = item.note
        return self.note

    async def delete(self) -> None:
        """|coro|

        Delete this item.
        """
        await self._state.delete_list_item(self.channel.id, self.id)

    async def edit(
        self,
        *pos_message: Optional[Union[str, Embed, File, Emoji]],
        message: Optional[str] = MISSING,
        note_content: Optional[str] = MISSING,
        position: int = None,
    ) -> ListItem:
        """|coro|

        Edit this item.

        .. note::

            If ``position`` and ``message`` are both specified, this method
            will make multiple API requests, unless the client is a bot account,
            in which case the ``position`` parameter is silently ignored.

        All parameters are optional.

        Parameters
        -----------
        \*pos_message: Union[:class:`str`, :class:`.Embed`, :class:`.File`, :class:`.Emoji`]
            An argument list of the list's main message, passed in the order
            that each element should display.
            If the client is a bot account, only the first value is used as content.
            This parameter cannot be combined with ``message``.
        message: :class:`str`
            The text content to set as the main message of the item.
            This parameter exists so that text content can be passed using a keyword argument.
            This parameter cannot be combined with ``pos_message``.
        note_content: :class:`str`
            The item's note content.
        position: :class:`int`
            The new position of the item. A value of ``0`` appears at the
            bottom visually.

        Returns
        --------
        :class:`.ListItem`
            The newly edited item.
            If the client is a user account and only the ``priority`` parameter
            was provided, this is the current instance rather than a new one.
        """

        # The parameters are named ``message`` for API compliance but we use ``content`` internally for consistency.
        content, pos_content = message, pos_message

        if content is not MISSING and pos_content:
            raise ValueError('Cannot provide both message and pos_message')

        if self._state.userbot:
            if position is not None:
                if not isinstance(position, int):
                    raise TypeError(f'position must be type int, not {position.__class__.__name__}')

                rich_positions = []
                all_items = await self.channel.fetch_items()
                for item in all_items:
                    rich_positions.append(item)

                rich_positions.sort(key=lambda item: item.position)
                rich_positions.insert(position, self)

                positions = [item.id for item in rich_positions]
                await self._state.edit_list_item_priorities(self.channel.id, positions)
                self.position = position

            if content is not MISSING or pos_content:
                if content is not MISSING:
                    pos_content = (content,)

                payload = {}
                if pos_content:
                    payload['message'] = self._state.compatible_content(await self._state.process_list_content(pos_content))

                if note_content is not MISSING:
                    if note_content is None:
                        payload['note'] = None
                    else:
                        payload['note'] = self._state.compatible_content([note_content])

                data = await self._state.edit_list_item_message(self.channel.id, self.id, payload=payload)
                item = ListItem(data=data, channel=self, state=self._state)
                return item

            return self

        else:
            if pos_content:
                content = pos_content[0]

            payload = {}
            if content is not MISSING:
                payload['message'] = content
            else:
                payload['message'] = self.message

            if note_content is not MISSING and note_content is not None:
                payload['note'] = {
                    'content': note_content
                }

            data = await self._state.update_list_item(
                self.channel_id,
                self.id,
                payload=payload,
            )
            item = ListItem(data=data['listItem'], channel=self, state=self._state)
            return item

    async def create_item(
        self,
        *pos_message: Optional[Union[str, Embed, File, Emoji]],
        message: str = MISSING,
        note_content: Optional[str] = MISSING,
        position: Optional[int] = None,
    ) -> ListItem:
        """|coro|

        |onlyuserbot|

        Create an item with this item as its parent.

        This method is identical to :meth:`ListChannel.create_item`.
        """
        return await self.channel.create_item(
            *pos_message,
            message=message,
            note_content=note_content,
            parent=self,
            position=position,
        )

    async def move(self, to: ListChannel) -> None:
        """|coro|

        |onlyuserbot|

        Move this item to another channel.

        Parameters
        -----------
        to: :class:`.ListChannel`
            The list channel to move this item to.
        """
        await self._state.move_list_item(self.channel.id, self.id, to.id)

    async def complete(self) -> None:
        """|coro|

        Mark this list item as complete.

        If this item has any children, they will also be marked as complete.
        """
        if self._state.userbot:
            await self._state.set_list_item_complete(self.channel.id, self.id, True)
        else:
            await self._state.complete_list_item(self.channel.id, self.id)

    async def uncomplete(self) -> None:
        """|coro|

        Mark this list item as incomplete.
        """
        if self._state.userbot:
            await self._state.set_list_item_complete(self.channel.id, self.id, False)
        else:
            await self._state.uncomplete_list_item(self.channel.id, self.id)


class MediaChannel(guilded.abc.TeamChannel):
    """Represents a media channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        self.type = ChannelType.media
        self.content_type = 'team_media'
        self._medias = {}

    @property
    def medias(self) -> List[Media]:
        """List[:class:`.Media`]: The list of cached medias in this channel."""
        return list(self._medias.values())

    def get_media(self, id) -> Optional[Media]:
        """Optional[:class:`.Media`]: Get a cached media post in this channel."""
        return self._medias.get(id)

    async def getch_media(self, id: int) -> Media:
        return self.get_media(id) or await self.fetch_media(id)

    async def fetch_media(self, id: int) -> Media:
        """|coro|

        |onlyuserbot|

        Fetch a media post in this channel.

        Parameters
        -----------
        id: :class:`int`
            The media's ID.

        Returns
        --------
        :class:`.Media`
        """
        data = await self._state.get_media(self.id, id)
        media = Media(data=data, channel=self, state=self._state)
        return media

    async def fetch_medias(self, *, limit: int = 50) -> List[Media]:
        """|coro|

        |onlyuserbot|

        Fetch multiple media posts in this channel.

        All parameters are optional.

        Parameters
        -----------
        limit: :class:`int`
            The maximum number of media posts to return. Defaults to 50.

        Returns
        --------
        List[:class:`.Media`]
        """
        data = await self._state.get_medias(self.id, limit=limit)
        medias = []
        for media_data in data:
            medias.append(Media(data=media_data, channel=self, state=self._state))

        return medias

    async def create_media(
        self,
        *,
        title: str,
        description: str = None,
        file: Optional[File] = None,
        youtube_url: Optional[str] = None,
        tags: List[str] = None,
        game: Optional[Game] = None,
    ) -> Media:
        """|coro|

        |onlyuserbot|

        Create a media post in this channel.

        Parameters
        -----------
        title: :class:`str`
            The title of the media.
        description: Optional[:class:`str`]
            The description of the media. Does not accept markdown or any
            inline content.
        file: :class:`.File`
            The file to upload. Either this or ``youtube_url`` is required.
        youtube_url: :class:`str`
            The YouTube embed URL to use (``https://www.youtube.com/embed/...``).
            Either this or ``file`` is required.
        game: Optional[:class:`.Game`]
            The game associated with the media.
        tags: List[:class:`str`]
            The tags on the media.

        Returns
        --------
        :class:`.Media`
            The created media.
        """
        if file and youtube_url:
            raise ValueError('Must not specify both file and youtube_url')
        if not file and not youtube_url:
            raise ValueError('Must specify either file or youtube_url')

        payload = {
            'title': title,
            'description': description or '',
            'tags': tags or [],
        }

        if file:
            file.set_media_type(MediaType.media_channel_upload)
            await file._upload(self._state)
            payload['src'] = file.url
            payload['type'] = file.file_type.value
        elif youtube_url:
            data = await self._state.upload_third_party_media(youtube_url)
            payload['src'] = data['url']
            payload['additionalInfo'] = {'externalVideoSrc': youtube_url}
            payload['type'] = FileType.video.value

        if game is not None:
            payload['gameId'] = game.id

        data = await self._state.create_media(self.id, payload=payload)
        media = Media(data=data, channel=self, game=game, state=self._state)
        return media


class ListChannel(guilded.abc.TeamChannel):
    """Represents a list channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        self.type = ChannelType.list
        self._items = {}

    @property
    def items(self) -> List[ListItem]:
        """List[:class:`.ListItem`]: The list of cached items in this channel."""
        return list(self._items.values())

    def get_item(self, id) -> Optional[ListItem]:
        """Optional[:class:`.ListItem`]: Get a cached item in this channel."""
        return self._items.get(id)

    async def getch_item(self, id: str) -> ListItem:
        return self.get_item(id) or await self.fetch_item(id)

    async def fetch_item(self, id: str) -> ListItem:
        """|coro|

        Fetch a item in this channel.

        Parameters
        -----------
        id: :class:`str`
            The item's ID.

        Returns
        --------
        :class:`.ListItem`
        """
        data = await self._state.get_list_item(self.id, id)
        if not self._state.userbot:
            data = data['listItem']
        item = ListItem(data=data, channel=self, state=self._state)
        return item

    async def fetch_items(self) -> List[ListItem]:
        """|coro|

        Fetch all items in this channel.

        Returns
        --------
        List[:class:`.ListItem`]
        """
        data = await self._state.get_list_items(self.id)
        if not self._state.userbot:
            data = data['listItems']
        items = []
        for item_data in data:
            items.append(ListItem(data=item_data, channel=self, state=self._state))

        return items

    async def create_item(
        self,
        *pos_message: Optional[Union[str, Embed, File, Emoji]],
        message: str = MISSING,
        note_content: Optional[str] = None,
        parent: Optional[ListItem] = None,
        position: Optional[int] = None,
    ) -> ListItem:
        """|coro|

        Create an item in this channel.

        Parameters
        -----------
        \*pos_message: Union[:class:`str`, :class:`.Embed`, :class:`.File`, :class:`.Emoji`]
            An argument list of the list's main message, passed in the order
            that each element should display.
            If the client is a bot account, only the first value is used as content.
            This parameter cannot be combined with ``message``.
        message: :class:`str`
            The text content to include as the main message of the item.
            This parameter exists so that text content can be passed using a keyword argument.
            This parameter cannot be combined with ``pos_message``.
        note_content: Optional[:class:`str`]
            The item's note content.
        parent: Optional[:class:`ListItem`]
            An existing item to create this item under.
        position: Optional[:class:`int`]
            The item's position. Defaults to ``0`` if not specified (appears
            at the bottom of the list).

        Returns
        --------
        :class:`.ListItem`
            The created item.
        """

        # The parameters are named ``message`` for API compliance but we use ``content`` internally for consistency.
        content, pos_content = message, pos_message

        if content is not MISSING and pos_content:
            raise ValueError('Cannot provide both message and pos_message')

        if self._state.userbot:
            if content is not MISSING:
                pos_content = (content,)

            data = await self._state.create_list_item(
                self.id,
                message=await self._state.process_list_content(pos_content),
                note_content=note_content,
                parent_id=parent.id if parent is not None else None,
                position=position if position is not None else 0,
            )

        else:
            if pos_content:
                content = pos_content[0]

            data = await self._state.create_list_item(
                self.id,
                message=content,
                note_content=note_content,
            )
            data = data['listItem']

        item = ListItem(data=data, channel=self, state=self._state)
        return item

    async def create_webhook(self, *, name: str, avatar: Optional[Union[bytes, File]] = None) -> Webhook:
        """|coro|

        Create a webhook in this channel.

        Parameters
        -----------
        name: :class:`str`
            The webhook's name.
        avatar: Optional[Union[:class:`bytes`, :class:`File`]]
            A :term:`py:bytes-like object` or :class:`File` for the webhook's avatar.
            If the client is a bot user, providing this does nothing.
            Else, providing this causes the library to perform an extra API request.

        Returns
        --------
        :class:`Webhook`
            The created webhook.

        Raises
        -------
        HTTPException
            Creating the webhook failed.
        Forbidden
            You do not have permissions to create a webhook.
        """

        webhook = await self.team.create_webhook(
            channel=self,
            name=name,
            avatar=avatar,
        )
        return webhook

    async def webhooks(self) -> List[Webhook]:
        """|coro|

        Fetch the list of webhooks in this channel.

        .. warning::

            This endpoint cannot be paginated.

        Returns
        --------
        List[:class:`Webhook`]
            The webhooks in this channel.

        Raises
        -------
        Forbidden
            You do not have permissions to get the webhooks.
        """

        webhooks = await self.team.webhooks(channel=self)
        return webhooks


class Availability:
    """Represents an availability in a :class:`.SchedulingChannel`\.

    Attributes
    -----------
    id: :class:`int`
        The availability's ID.
    start: :class:`datetime.datetime`
        When the availability starts.
    end: :class:`datetime.datetime`
        When the availability ends.
    created_at: :class:`datetime.datetime`
        When the availabilty was created.
    updated_at: Optional[:class:`datetime.datetime`]
        When the availabilty was updated.
    deleted_by: Optional[:class:`.Member`]
        The member that deleted the availability.
    channel: :class:`.SchedulingChannel`
        The channel that the availability is in.
    """
    def __init__(self, *, state, data, channel):
        self._state = state

        self.channel: SchedulingChannel = channel
        self.channel_id: str = data.get('channelId')
        self.team_id: str = data.get('teamId')
        self.user_id: str = data.get('userId')

        self.id: int = data.get('id', data.get('availabilityId'))
        self.created_at: datetime.datetime = ISO8601(data.get('createdAt'))
        self.updated_at: Optional[datetime.datetime] = ISO8601(data.get('updatedAt'))
        self.start: datetime.datetime = ISO8601(data.get('startDate'))
        self.end: datetime.datetime = ISO8601(data.get('endDate'))

        self.updated_by_id: Optional[str] = data.get('updatedBy')
        self.deleted_by: Optional[Member] = None

    def __eq__(self, other) -> bool:
        return isinstance(other, Availability) and other.id == self.id

    def __repr__(self) -> str:
        return f'<Availability id={self.id!r} start={self.start!r} end={self.end!r} channel={self.channel!r}>'

    @property
    def team(self):
        """:class:`.Team`: The team that the availability is in."""
        return self.channel.team

    @property
    def group(self) -> Optional[Group]:
        """Optional[:class:`.Group`]: The group that the availability is in."""
        return self.channel.group

    @property
    def updated_by(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The member that last updated the availability."""
        return self.team.get_member(self.updated_by_id)

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`.Member`]: The member that created the availability."""
        return self.team.get_member(self.user_id)

    async def edit(self, *,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
    ) -> Availability:
        """|coro|

        |onlyuserbot|

        Edit this availability.

        All parameters are optional.

        Parameters
        -----------
        start: :class:`datetime.datetime`
            When the availability starts.
            Time must be in a 30-minute interval (``minute`` must be 0 or 30).
        end: :class:`datetime.datetime`
            When the availability ends.
            Time must be in a 30-minute interval (``minute`` must be 0 or 30).

        Returns
        --------
        :class:`.Availability`
            The newly edited availability.
        """
        payload = {
            'startDate': start or self.start,
            'endDate': end or self.end,
        }
        data = await self._state.update_availability(self.channel.id, self.id, payload=payload)
        self.start = payload['startDate']
        self.end = payload['endDate']
        return self

    async def delete(self) -> None:
        """|coro|

        |onlyuserbot|

        Delete this availability.
        """
        await self._state.delete_availability(self.channel.id, self.id)


class SchedulingChannel(guilded.abc.TeamChannel):
    """Represents a scheduling channel in a :class:`.Team`\."""
    def __init__(self, **fields):
        super().__init__(**fields)
        self.type = ChannelType.scheduling
        self._availabilities = {}

    @property
    def availabilities(self) -> List[Availability]:
        """List[:class:`.Availability`]: The list of cached availabilities in this channel."""
        return list(self._availabilities.values())

    def get_availability(self, id) -> Optional[Availability]:
        """Optional[:class:`.Availability`]: Get a cached availability in this channel."""
        return self._availabilities.get(id)

    async def getch_availability(self, id: str) -> Availability:
        return self.get_availability(id) or await self.fetch_availability(id)

    async def fetch_availability(self, id: int) -> Availability:
        """|coro|

        Fetch an availability in this channel.

        .. note::

            There is no endpoint to fetch a specific availability, so instead
            this method filters :meth:`.fetch_availabilities` and raises
            :exc:`InvalidArgument` if no availability was found.

        Parameters
        -----------
        id: :class:`int`
            The availability's ID.

        Returns
        --------
        :class:`.Availability`

        Raises
        -------
        InvalidArgument
            No availability exists in this channel with the ID specified.
        """
        availabilities = await self.fetch_availabilities()
        availability = get(availabilities, id=id)
        if not availability:
            raise InvalidArgument(f'No availability exists in this channel with the ID {id}.')
        return availability

    async def fetch_availabilities(self) -> List[Availability]:
        """|coro|

        Fetch all availabilities in this channel.

        Returns
        --------
        List[:class:`.Availability`]
        """
        data = await self._state.get_availabilities(self.id)
        availabilities = []
        for availability_data in data:
            availabilities.append(Availability(data=availability_data, channel=self, state=self._state))

        return availabilities

    async def create_availability(self, start: datetime.datetime, end: datetime.datetime) -> Availability:
        """|coro|

        Create an availability in this channel.

        Parameters
        -----------
        start: :class:`datetime.datetime`
            When the availability starts.
            Time must be in a 30-minute interval (``minute`` must be 0 or 30).
        end: :class:`datetime.datetime`
            When the availability ends.
            Time must be in a 30-minute interval (``minute`` must be 0 or 30).

        Returns
        --------
        :class:`.Availability`
            The created availability.
        """
        data = await self._state.create_availability(self.id, start=start, end=end)
        for availability_data in data['availabilities']:
            availability = Availability(data=availability_data, channel=self, state=self._state)
            if availability.id == data['id']:
                return availability


class AnnouncementReply(guilded.abc.Reply):
    """Represents a reply to an :class:`Announcement`\."""
    pass


class DocReply(guilded.abc.Reply):
    """Represents a reply to a :class:`Doc`\."""
    pass


class ForumReply(guilded.abc.Reply):
    """Represents a reply to a :class:`ForumTopic`\."""
    pass


class MediaReply(guilded.abc.Reply):
    """Represents a reply to a :class:`Media`\."""
    pass
